/*
        SHELL.C         a simple user-extendable command 
                        interpreter for MS-DOS 2.X and 3.X

        Copyright (C) 1985 Ray Duncan

        To compile with Microsoft C 3.0 and 
        link into the executable file SHELL.EXE:
 
                        C>MSC SHELL;
                        C>LINK SHELL;
*/

#include <stdio.h>
#include <process.h>
#include <stdlib.h>
#include <signal.h>

#define dim(x) (sizeof(x)/sizeof(x[0]))	/* macro to return the number of
					   elements in a structure */

int break_handler();			/* handler routine for Control-C */

int cls_cmd(), dos_cmd(), exit_cmd();	/* declare intrinsic processors
					   for use in command table */

struct cmd_table {			/* table of intrinsic commands */
           char *cmd_name;              /* command name entered by user */
           int  (*cmd_fxn)();           /* corresponding fxn to execute */      
           }   commands[] =
               {   "CLS",   cls_cmd,
                   "DOS",   dos_cmd,
                   "EXIT",  exit_cmd,
               };

static char com_spec[64];		/* filespec of COMMAND.COM from
					   Environment Block placed here */


main(argc, argv)
int   argc;
char  *argv[];

{   char inp_buf[80];                	/* operator's command placed here */ 

    get_comspec(com_spec);		/* get filespec for COMMAND.COM */

					/* take over Break Interrupt 23H
					   so shell won't lose control */	
    if (signal(SIGINT,break_handler) == (int(*)()) -1)
        {   fputs("Can't capture Control-C Interrupt",stderr);
	    exit(1); 
        }

    while(1)                            /* main command interpreter loop */
        {   get_cmd(inp_buf);           /* get a command */
            if (! intrinsic(inp_buf) )	/* if it's an intrinsic command,
                                           run its subroutine */
                extrinsic(inp_buf);	/* else pass it to COMMAND.COM */
        }
}


/*      
        Try and match user's command with intrinsic command table.
        If a match is found, run the associated routine and return
        a True flag, else return a False flag.
*/

intrinsic(input_string)
char *input_string;
{   int i, j;
					/* scan off leading blanks */
    while( *input_string == '\x20') input_string++ ;
					/* search the command table */
    for (i=0; i < dim(commands); i++)
        {   j = strcmp( commands[i].cmd_name,input_string );
            if  (j == 0) 		/* if match found, run routine */
                {   ( *commands[i].cmd_fxn )() ;  
                    return(1);		/* and return a True flag */
                }
        }
    return(0);				/* no match, return False flag */
}


/* 
        Process an extrinsic command by passing it
        to an EXEC'd copy of COMMAND.COM.
*/

extrinsic(input_string)
char *input_string;
{   int status;
    status = system(input_string);	/* call EXEC function */
    if (status) 			/* if failed, print error message */
	fputs("\nEXEC of COMMAND.COM failed\n",stderr);
}


/* 
        Issue prompt, get command line from the Standard Input Device
        as a null-terminated (ASCIIZ) string, fold it to upper case.
*/

get_cmd(buffer)
char *buffer;
{   printf("\nsh: ");                   /* display prompt */
    gets(buffer);                       /* get line from Standard Input */
    strupr(buffer);                     /* fold it to upper case */
}       


/*	
	Get the full path and file specification for COMMAND.COM
	from the "COMSPEC=" variable in the Environment Block
*/

get_comspec(buffer)
char *buffer;
{   strcpy( buffer, getenv("COMSPEC") );
    if  ( buffer[0] == NULL )		
        {   fputs("\nNo COMSPEC variable in Environment\n",stderr);
 	    exit(1);
        }
}


/*
	This handler for Interrupt 23H signal keeps our shell from 
	losing control when Control-C is entered.  Just re-issues
	the prompt and returns.
*/

break_handler()
{   signal(SIGINT,break_handler);	/* reset handler address */
    printf("\nsh: ");                   /* display prompt */
}	


/*      
        These are the subroutines for the various intrinsic commands.
*/

cls_cmd()               /* routine for intrinsic CLS command */
{   printf("\033[2J");  /* this is ANSI clear screen string */
}

dos_cmd()               /* routine for intrinsic DOS command */
{   int status;
    status = spawnlp(P_WAIT,com_spec,com_spec,NULL);
    if (status)	fputs("\nEXEC of COMMAND.COM failed\n",stderr); 
}

exit_cmd()              /* routine for intrinsic EXIT command */
{   exit(0);
}
